import subprocess
import os
import sys
import logging
import shutil
import xmlrpc.client

import requests
import jinja2

from .resource import load_resource


def _esc(s):
    return s.replace('\\', '\\\\').replace('"', '\\"')


class Downloader:
    def __init__(self, *, dl_tasks):
        self._dl_tasks = dl_tasks

    def download(self):
        pass


class DownloaderTS(Downloader):
    def __init__(self, *args, ts, **kwargs):
        super().__init__(*args, **kwargs)

        self._ts = ts
        self._dl = ts.register_task(self._dl, ttl=3, desc='download')

    def _dl(self, *, url, filename):
        pass

    def download(self):
        for _ in self._dl_tasks:
            self._dl(url=_['url'], filename=_['filename'])

        failures = self._ts.wait()
        return [_1.kwargs for _1, _2 in failures]


class DownloaderBuiltin(DownloaderTS):
    def _dl(self, *, url, filename):
        os.makedirs(os.path.dirname(filename), exist_ok=True)

        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            r.raw.decode_content = True
            with open(filename, 'wb') as ofs:
                shutil.copyfileobj(r.raw, ofs)


class DownloaderSubprocess(DownloaderTS):
    def _dl(self, *, url, filename):
        subprocess.run(self._cmdline(url, filename), stdout=subprocess.DEVNULL,
                       stderr=subprocess.STDOUT, check=True)

    def _cmdline(self, url, filename):
        pass


class DownloaderCurl(DownloaderSubprocess):
    def _cmdline(self, url, filename):
        return ['curl', '--create-dirs', '--globoff', '--url', url, '--output', filename]


class DownloaderAria2(DownloaderSubprocess):
    def _cmdline(self, url, filename):
        return ['aria2c',
                '--allow-overwrite=true',
                '--always-resume=true',
                '--auto-file-renaming=false',
                '--continue=true',
                '-d', os.path.dirname(filename),
                '-o', os.path.basename(filename),
                url]


class DownloaderInput_file(Downloader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._env = jinja2.Environment()
        self._env.filters['_dirname'] = lambda _: os.path.dirname(_)
        self._env.filters['_basename'] = lambda _: os.path.basename(_)
        self._env.filters['_esc'] = _esc

    def download(self):
        template = self._env.from_string(load_resource(self._template_path).decode('UTF-8'))
        return template.render(dl_tasks=self._dl_tasks)

    @property
    def _template_path(self):
        pass


class DownloaderCurl_input_file(DownloaderInput_file):
    @property
    def _template_path(self):
        return 'template/curl_input_file'


class DownloaderAria2_input_file(DownloaderInput_file):
    @property
    def _template_path(self):
        return 'template/aria2_input_file'


class DownloaderUget(Downloader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._uget = 'uget' if sys.platform == 'win32' else 'uget-gtk'

    def download(self):
        for _ in self._dl_tasks:
            _url = _['url']
            _filename = _['filename']
            cmdline = [self._uget, '--quiet',
                       '--folder=%s' % os.path.dirname(_filename),
                       '--filename=%s' % os.path.basename(_filename), _url]
            subprocess.run(cmdline, stdout=subprocess.DEVNULL,
                           stderr=subprocess.STDOUT, check=True)
            logging.info('add downloading task: url=%s, filename=%s' % (_url, _filename))


class DownloaderAria2_rpc(Downloader):
    def __init__(self, *args, url=None, secret=None, **kwargs):
        super().__init__(*args, **kwargs)

        self._client = xmlrpc.client.ServerProxy(url)
        self._secret = secret

    def download(self):
        for _ in self._dl_tasks:
            _url = _['url']
            _filename = _['filename']
            _args = [[_url], {'dir': os.path.dirname(_filename),
                              'out': os.path.basename(_filename),
                              'allow-overwrite': 'true',
                              'always-resume': 'true',
                              'auto-file-renaming': 'false',
                              'continue': 'true'}]
            if self._secret is not None:
                _args.insert(0, 'token:%s' % self._secret)

            self._client.aria2.addUri(*_args)
            logging.info('add downloading task: url=%s, filename=%s' % (_url, _filename))



import logging
import argparse
import os
import pickle
import json
import sys
import textwrap

import requests

import dl_coursera

from dl_coursera.lib.misc import change_ext
from dl_coursera.lib.TaskScheduler import TaskScheduler
from dl_coursera.Crawler import Crawler
from dl_coursera.DLTaskGatherer import DLTaskGatherer
from dl_coursera.Downloader import *


def _file_pkl_crawl(outdir, slug):
    return os.path.join(outdir, '%s.crawl.pkl' % slug)


def _file_json_gather(outdir, slug):
    return os.path.join(outdir, '%s.gather.json' % slug)


def _file_json_download_dl_tasks_failed(outdir, slug):
    return os.path.join(outdir, '%s.download.dl_tasks_failed.json' % slug)


def _file_txt_download_input_file(outdir, slug, how):
    return os.path.join(outdir, '%s.download.%s_input_file.txt' % (slug, how))


def crawl(email, password, slug, isSpec, outdir, n_worker):
    file_pkl = _file_pkl_crawl(outdir, slug)
    if os.path.exists(file_pkl):
        with open(file_pkl, 'rb') as ifs:
            return pickle.load(ifs)

    with TaskScheduler() as ts, requests.Session() as sess:
        ts.start(n_worker=n_worker)
        crawler = Crawler(ts=ts, sess=sess, email=email, password=password)
        soc = crawler.crawl(slug=slug, isSpec=isSpec)

    with open(file_pkl, 'wb') as ofs:
        pickle.dump(soc, ofs)

    file_json = change_ext(file_pkl, 'json')
    with open(file_json, 'w', encoding='UTF-8') as ofs:
        ofs.write(soc.to_json())

    return soc


def gather_dl_tasks(outdir, soc):
    file_json = _file_json_gather(outdir, soc['slug'])
    if os.path.exists(file_json):
        with open(file_json, encoding='UTF-8') as ifs:
            return json.load(ifs)

    dl_tasks = DLTaskGatherer(soc=soc, outdir=outdir).gather()
    with open(file_json, 'w', encoding='UTF-8') as ofs:
        json.dump(dl_tasks, ofs)

    return dl_tasks


def download_ts(dl_tasks, slug, outdir, how):
    file_json = _file_json_download_dl_tasks_failed(outdir, slug)
    if os.path.exists(file_json):
        with open(file_json, encoding='UTF-8') as ifs:
            dl_tasks = json.load(ifs)

    if len(dl_tasks) == 0:
        return

    with TaskScheduler() as ts:
        ts.start(n_worker=4)

        _cls_downloader = {'builtin': DownloaderBuiltin,
                           'curl': DownloaderCurl,
                           'aria2': DownloaderAria2}[how]
        dl_tasks_failed = _cls_downloader(dl_tasks=dl_tasks, ts=ts).download()

    with open(file_json, 'w', encoding='UTF-8') as ofs:
        json.dump(dl_tasks_failed, ofs)


def download_input_file(dl_tasks, slug, outdir, how):
    _cls_downloader = {'curl': DownloaderCurl_input_file,
                       'aria2': DownloaderAria2_input_file}[how]
    s = _cls_downloader(dl_tasks=dl_tasks).download()

    file_txt = _file_txt_download_input_file(outdir, slug, how)
    with open(file_txt, 'w', encoding='UTF-8') as ofs:
        ofs.write(s)
    logging.info('new file: %s' % file_txt)


def main():
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s - %(threadName)s - %(message)s')

    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=True,
        description="A simple, fast, and reliable Coursera crawling & downloading tool",
        epilog=textwrap.dedent('''
            If the command succeeds, you shall see `Done :-)'.
            If some UNEXPECTED errors occur, try deleting everything generated by this
            tool in @outdir, and then run the command again.
            For more information, visit `https://github.com/feng-lei/dl_coursera'.
            '''
        )
    )
    parser.add_argument('--version', action='version', version='%%(prog)s %s' % dl_coursera.app_version)
    parser.add_argument('--email', required=True)
    parser.add_argument('--password', required=True)
    parser.add_argument('--slug', required=True,
        help='slug of a course or a specializtion (with @--isSpec)')
    parser.add_argument('--isSpec', action='store_true',
        help='indicate that @slug is slug of a specialization')
    parser.add_argument('--n-worker', type=int, default=4,
        help='''the number of threads used to crawl webpages. Default: 4.
                NOTE: if errors show up during crawling, try decreasing this value''')

    parser.add_argument('--outdir', default='.',
        help='the directory to save files to. Default: `.\'')
    parser.add_argument('--how', required=True,
        choices=['builtin', 'curl', 'aria2', 'aria2-rpc', 'uget'],
        help='''how to download files.
                builtin (NOT recommonded): use the builtin downloader.
                curl: invoke the `curl' tool or generate an "input file" for that
                      tool (with @--generate-input-file).
                aria2: invoke the `aria2c' tool or generate an "input file" for that
                      tool (with @--generate-input-file).
                aria2-rpc (HIGHLY recommonded): add downloading tasks to aria2
                      through its XML-RPC interface.
                uget (recommonded): add downloading tasks to the uGet Download Manager'''
    )
    parser.add_argument('--generate-input-file', action='store_true',
        help='''when @--how is curl/aria2, indicate that to generate an "input file"
                for that tool, rather than to invoke it''')
    parser.add_argument('--aria2-rpc-url', default='http://localhost:6800/rpc',
        help="url of the aria2 XML-RPC interface. Default: `http://localhost:6800/rpc'")
    parser.add_argument('--aria2-rpc-secret', help='authorization token of the aria2 XML-RPC interface')

    args = vars(parser.parse_args())

    os.makedirs(args['outdir'], exist_ok=True)

    soc = crawl(args['email'], args['password'], args['slug'],
                args['isSpec'], args['outdir'], args['n_worker'])

    dl_tasks = gather_dl_tasks(args['outdir'], soc)

    if args['how'] == 'builtin':
        download_ts(dl_tasks, args['slug'], args['outdir'], args['how'])

    elif args['how'] in ['curl', 'aria2']:
        if args['generate_input_file']:
            download_input_file(dl_tasks, args['slug'], args['outdir'], args['how'])
        else:
            download_ts(dl_tasks, args['slug'], args['outdir'], args['how'])

    elif args['how'] == 'aria2-rpc':
        DownloaderAria2_rpc(dl_tasks=dl_tasks,
                            url=args['aria2_rpc_url'],
                            secret=args['aria2_rpc_secret']).download()

    elif args['how'] == 'uget':
        DownloaderUget(dl_tasks=dl_tasks).download()

    print('\nDone :-)')


if __name__ == '__main__':
    main()



import pickle
import os
import logging

from .define import *

from .lib.misc import format_dict
from .markup import CML

PRIO_SPEC = 'A'
PRIO_COURSE = 'B'
PRIO_COURSE_MATERIAL = 'C'


class Crawler:
    def __init__(self, *, ts, sess, email, password):
        self._ts = ts

        def attach(func):
            setattr(self, func.__name__, func)
            return func

        @attach
        def login():
            resp = sess.get(URL_ROOT)
            resp = sess.post(URL_LOGIN(resp.cookies['CSRF3-Token']),
                             data={'email': email, 'password': password})

            assert resp.status_code == 200

        @attach
        @ts.register_task(
            priority=PRIO_SPEC, ttl=3,
            format_kwargs=lambda _: format_dict({'spec': _['spec']['slug']})
        )
        def crawl_spec(*, spec):
            resp = sess.get(URL_SPEC(spec['slug']))
            d = resp.json()

            _ = d['elements'][0]
            spec['id'] = _['id']
            spec['name'] = _['name']

            assert spec['slug'] == _['slug']

            for id_ in _['courseIds']:
                spec['courses'].append(Course(id_=id_))

            id2slug = {}
            for _ in d['linked']['courses.v1']:
                id2slug[_['id']] = _['slug']

            for _ in spec['courses']:
                _['slug'] = id2slug[_['id']]

            logging.info('Courses of specialization %s: %s' % (
                spec['slug'],
                ', '.join([_['slug'] for _ in spec['courses']])
            ))

            for _ in spec['courses']:
                crawl_course(course=_)

        @attach
        @ts.register_task(
            priority=PRIO_COURSE, ttl=3,
            format_kwargs=lambda _: format_dict({'source': _['course']['slug']})
        )
        def crawl_course(*, course):
            resp = sess.get(URL_COURSE_1(course['slug']))
            d = resp.json()

            _ = d['elements'][0]
            course['name'] = _['name']
            course['id'] = _['id']

            assert course['slug'] == _['slug']

            resp = sess.get(URL_COURSE_2(course['slug']))
            d = resp.json()['linked']

            id2item = {}
            for _ in d['onDemandCourseMaterialItems.v2']:
                typeName = _['contentSummary']['typeName']
                if typeName == 'lecture':
                    id2item[_['id']] = CourseMaterialLecture(id_=_['id'], name=_['name'], slug=_['slug'])
                elif typeName == 'notebook':
                    id2item[_['id']] = CourseMaterialNotebook(id_=_['id'], name=_['name'], slug=_['slug'])
                elif typeName == 'supplement':
                    id2item[_['id']] = CourseMaterialSupplement(id_=_['id'], name=_['name'], slug=_['slug'])

            id2lesson = {}
            for _ in d['onDemandCourseMaterialLessons.v1']:
                lesson = CourseMaterialLesson(id_=_['id'], name=_['name'], slug=_['slug'])
                for id_item in _['itemIds']:
                    item = id2item.get(id_item)
                    if item is not None:
                        lesson['items'].append(item)

                if len(lesson['items']) > 0:
                    id2lesson[lesson['id']] = lesson

            for _ in d['onDemandCourseMaterialModules.v1']:
                module = CourseMaterialModule(id_=_['id'], name=_['name'], slug=_['slug'])

                for id_ in _['lessonIds']:
                    lesson = id2lesson.get(id_)
                    if lesson is not None:
                        module['lessons'].append(lesson)

                if len(module['lessons']) > 0:
                    course['modules'].append(module)

            for module in course['modules']:
                for lesson in module['lessons']:
                    for item in lesson['items']:
                        if item['type'] == 'Lecture':
                            crawl_lecture(course=course, lecture=item)
                        elif item['type'] == 'Supplement':
                            crawl_supplement(course=course, supplement=item)

        @ts.register_task(
            priority=PRIO_COURSE_MATERIAL, ttl=3,
            format_kwargs=lambda _: format_dict({'course': _['course']['slug'],
                                                 'lecture': _['lecture']['slug']})
        )
        def crawl_lecture(*, course, lecture):
            # lecture videos
            resp = sess.get(URL_LECTURE_1(course['id'], lecture['id']))
            d = resp.json()

            for _ in d['linked']['onDemandVideos.v1']:
                url_subtitle = _['subtitles'].get('en')
                if url_subtitle is not None:
                    url_subtitle = URL_ROOT + url_subtitle

                url_video = None
                for reso in ['720p', '540p', '360p']:
                    url_video = _['sources']['byResolution'].get(reso)
                    if url_video is not None:
                        break

                assert url_video is not None

                url_video = url_video['mp4VideoUrl']
                lecture['videos'].append(Video(url_video=url_video, url_subtitle=url_subtitle))

            # lecture assets
            resp = sess.get(URL_LECTURE_2(course['id'], lecture['id']))
            d = resp.json()

            assetIDs = []
            for _ in d['linked']['openCourseAssets.v1']:
                typeName = _['typeName']
                if typeName == 'asset':
                    assetIDs.append(_['definition']['assetId'])
            assets = crawl_assets(assetIDs)
            lecture['assets'] = assets

        @ts.register_task(
            priority=PRIO_COURSE_MATERIAL, ttl=3,
            format_kwargs=lambda _: format_dict({'course': _['course']['slug'],
                                                 'supplement': _['supplement']['slug']})
        )
        def crawl_supplement(course, supplement):
            resp = sess.get(URL_SUPPLEMENT(course['id'], supplement['id']))
            d = resp.json()

            for _ in d['linked']['openCourseAssets.v1']:
                typeName = _['typeName']
                if typeName == 'cml':
                    cml = CML(_['definition']['value'])
                    assetIDs = cml.get_assetIDs()
                    assets = crawl_assets(assetIDs)
                    html = cml.to_html(assets=assets)

                    supplement['items'].append(CourseMaterialSupplementItemCML(html=html, assets=assets))

        def crawl_assets(ids):
            if len(ids) == 0:
                return []

            resp = sess.get(URL_ASSET(ids))
            d = resp.json()

            assets = []
            for _ in d['elements']:
                id_ = _['id']
                url = _['url']['url']
                name = _asset_name(_['name'], _['fileExtension'])

                assets.append(Asset(id_=id_, url=url, name=name))

            assert len(assets) == len(ids)
            return assets

        def _asset_name(name, fileExtension):
            fileExtension = '.' + fileExtension
            if not name.endswith(fileExtension):
                name += fileExtension
            return name

    def crawl(self, *, slug, isSpec):
        self.login()
        if isSpec:
            res = Spec(slug=slug)
            self.crawl_spec(spec=res)
        else:
            res = Course(slug=slug)
            self.crawl_course(course=res)
        failures = self._ts.wait()

        assert len(failures) == 0
        return res



import os
import zipfile
import io
import logging

from .lib.ExploringTree import ExploringTree

from .markup import render_supplement

from .resource import load_resource


def _shorten_slug(x):
    if len(x['slug']) > 40:
        x['slug'] = x['slug'][:40]


class DLTaskGatherer:
    def __init__(self, *, soc, outdir):  # "soc" means "sepc or course"
        assert soc['type'] in ['Spec', 'Course']

        self._soc = soc
        self._outdir = outdir

        self._et = ExploringTree()
        with self._et:
            self._resource_node = self._et.jump('%s/resource' % self._soc['slug'])

        self._dl_tasks = []
        self._file_tasks = []

    def gather(self):
        (self._gather_spec if self._soc['type'] == 'Spec' else self._gather_course) (self._soc)

        _dir = self._path(self._resource_node.abspath()[1:])
        if not os.path.exists(_dir):
            with zipfile.ZipFile(io.BytesIO(load_resource('bundle.zip'))) as zf:
                zf.extractall(_dir)
            logging.info('new directory tree: %s' % _dir)

        for _ in self._file_tasks:
            data = _['data']
            filename = _['filename']
            if not os.path.exists(filename):
                os.makedirs(os.path.dirname(filename), exist_ok=True)
                with open(filename, 'wb') as ofs:
                    ofs.write(data)
                logging.info('new file: %s' % filename)

        return self._dl_tasks

    def _resource_path(self):
        return self._et.relpathTo(self._resource_node)

    def _path(self, s):
        return os.path.abspath(os.path.join(self._outdir, s))

    def _add_dl_task(self, url, s):
        self._dl_tasks.append({'url': url, 'filename': self._path(s)})

    def _add_file_task(self, data, s):
        self._file_tasks.append({'data': data, 'filename': self._path(s)})

    def _down(self, s, i=None):
        if i is None:
            self._et.down(s)
        else:
            self._et.down('%02d@%s' % (i + 1, s))

    def _see(self, s):
        return self._et.see(s).abspath()[1:]

    def _gather_spec(self, spec):
        with self._et:
            self._down(spec['slug'])

            for _i, course in enumerate(spec['courses']):
                self._gather_course(course, _i)

    def _gather_course(self, course, i=None):
        with self._et:
            self._down(course['slug'], i)

            for _i, module in enumerate(course['modules']):
                self._gather_module(module, _i)

    def _gather_module(self, module, i):
        _shorten_slug(module)

        with self._et:
            self._down(module['slug'], i)

            for _i, lesson in enumerate(module['lessons']):
                self._gather_lesson(lesson, _i)

    def _gather_lesson(self, lesson, i):
        _shorten_slug(lesson)

        with self._et:
            self._down(lesson['slug'], i)

            for _i, item in enumerate(lesson['items']):
                if item['type'] == 'Lecture':
                    self._gather_lecture(item, _i)
                elif item['type'] == 'Supplement':
                    self._gather_supplement(item, _i)

    def _gather_lecture(self, lecture, i):
        _shorten_slug(lecture)

        with self._et:
            self._down(lecture['slug'], i)

            for _i, video in enumerate(lecture['videos']):
                self._gather_video(video, _i)

            for asset in lecture['assets']:
                self._gather_asset(asset)

    def _gather_video(self, video, i):
        self._add_dl_task(video['url_video'], self._see('%02d@.mp4' % (i + 1)))
        if video.get('url_subtitle') is not None:
            self._add_dl_task(video['url_subtitle'], self._see('%02d@.srt' % (i + 1)))

    def _gather_supplement(self, supplement, i):
        _shorten_slug(supplement)

        with self._et:
            self._down(supplement['slug'], i)

            for _i, item in enumerate(supplement['items']):
                if item['type'] == 'CML':
                    self._gather_cml(item, _i, supplement)

    def _gather_cml(self, cml, i, supplement):
        _data = render_supplement(content=cml['html'],
                                  resource_path=self._resource_path(),
                                  title='%s' % supplement['name']).encode('UTF-8')
        self._add_file_task(_data, self._see('%02d@.html' % (i + 1)))

        for asset in cml['assets']:
            self._gather_asset(asset)

    def _gather_asset(self, asset):
        self._add_dl_task(asset['url'], self._see(asset['name']))



import os


def format_dict(d):
    return ', '.join(['%s=%s' % (k, v) for k, v in d.items()])


def change_ext(filename, ext):
    return os.path.splitext(filename)[0] + ('' if ext == '' else '.' + ext)



import logging
import copy

import bs4

import jinja2

from .resource import load_resource


def _is_root(e):
    return e.parent is None


def _is_tag(e):
    return isinstance(e, bs4.Tag)


def _has_no_child(tag):
    return len(tag.contents) == 0


class Traversal:
    '''depth-first traversal
    '''

    def __init__(self, root, *, tagOnly=False):
        assert _is_root(root)

        self._e = root
        self._skip_children = False
        self._tagOnly = tagOnly

    def __iter__(self):
        return self

    def __next__(self):
        e = self._next()
        if e is None:
            raise StopIteration('no more element')

        self._e = e
        self._skip_children = False
        return e

    def _next(self):
        if (not self._skip_children) and _is_tag(self._e):
            for e in self._e.children:
                if (not self._tagOnly) or _is_tag(e):
                    return e

        e = self._e
        while True:
            if _is_root(e):
                return None

            for _e in e.next_siblings:
                if (not self._tagOnly) or _is_tag(_e):
                    return _e

            e = e.parent

    def skip_children(self):
        self._skip_children = True


class CML:
    def __init__(self, doc):
        doc = doc.translate(doc.maketrans('\u21b5', ' '))
        self._root = bs4.BeautifulSoup(doc, 'lxml-xml')

        self._assetIDs = None
        self._html = None

    def get_assetIDs(self):
        if self._assetIDs is not None:
            return self._assetIDs

        self._assetIDs = []
        for e in Traversal(self._root, tagOnly=True):
            if e.name == 'asset':
                self._assetIDs.append(e['id'])
            elif e.name == 'img':
                self._assetIDs.append(e['assetId'])
        return self._assetIDs

    def to_html(self, *, assets):
        if self._html is not None:
            return self._html

        asset_by_id = {_['id']: _ for _ in assets}

        def _assetName(id_):
            return asset_by_id[id_]['name']

        html = bs4.BeautifulSoup('', 'lxml')
        d = {}

        def _add(e0, e1):
            parent1 = html
            _e = e0
            while _e is not None:
                if id(_e) in d:
                    parent1 = d[id(_e)]
                    break
                _e = _e.parent

            if (parent1 is html) and (not _is_tag(e0)):
                return

            if _is_tag(e0):
                d[id(e0)] = e1
            parent1.append(e1)

        tr = Traversal(self._root)
        for e0 in tr:
            if isinstance(e0, bs4.NavigableString):
                _li = str(e0).split('$$')
                hasMath = False
                for _ in _li:
                    if not hasMath:
                        _add(e0, _)
                    else:
                        _span = bs4.Tag(name='span')
                        _span['hasMath'] = 'true'
                        _span.append(_)
                        _add(e0, _span)

                    hasMath = not hasMath

                continue

            if not _is_tag(e0):
                continue

            if e0.name == 'asset':
                assert _has_no_child(e0)

                e1 = bs4.Tag(name='p')
                e1['class'] = 'asset'
                e1.append(_assetName(e0['id']))

            elif e0.name == 'img':
                assert _has_no_child(e0)

                e1 = bs4.Tag(name='img')
                e1['src'] = e1['alt'] = _assetName(e0['assetId'])

            elif e0.name == 'heading':
                e1 = bs4.Tag(name='h%d' % int(e0['level']))

            elif e0.name == 'text':
                e1 = bs4.Tag(name='p')

            elif e0.name == 'list':
                bulletType = e0['bulletType']
                if bulletType == 'numbers':
                    e1 = bs4.Tag(name='ol')
                    e1['type'] = '1'
                elif bulletType == 'bullets':
                    e1 = bs4.Tag(name='ul')
                else:
                    e1 = bs4.Tag(name='ul')
                    logging.warning('[CML] unknown bulletType=%s' % bulletType)

            elif e0.name == 'a':
                e1 = bs4.Tag(name='a')
                e1['href'] = e0['href']

            elif e0.name == 'code':
                e1 = bs4.Tag(name='pre')
                e1.append(copy.copy(e0))

                tr.skip_children()

            elif e0.name in ['li', 'strong', 'em', 'table', 'tr', 'td']:
                e1 = bs4.Tag(name=e0.name)

            else:
                continue

            _add(e0, e1)

        self._html = str(html)
        return self._html


def render_supplement(*, content, resource_path, title='', __={}):
    if __.get('template') is None:
        __['template'] = jinja2.Template(load_resource('template/supplement.html').decode('UTF-8'))
    return __['template'].render(content=content, resource_path=resource_path, title=title)



import math
import queue
import threading
import traceback
import logging
import functools
import re

from .misc import format_dict


class Task:
    LOWEST_PRIO = ''
    HIGHEST_PRIO = 'ZZZZZZZ'

    _re_priority = re.compile(r'^[A-Z]{1,6}$')

    class ShouldNotRun(Exception):
        pass

    def __init__(self, *, priority, ttl, _isSpecial=False):
        """
        @priority: should be a string that matches the following Python regular expression: `[A-Z]{1, 6}`
        @ttl: maximum times to try
        """

        self._priority = priority
        self._ttl = ttl

        if not _isSpecial:
            assert Task._re_priority.match(self._priority) is not None

    def __lt__(self, other):
        return self._priority > other._priority

    def should_run(self):
        return self._ttl > 0

    def run(self):
        if not self.should_run():
            raise Task.ShouldNotRun(self)
        else:
            self._ttl -= 1
            self.go()

    def go():
        pass


_shutdown_task = Task(priority=Task.HIGHEST_PRIO, ttl=1, _isSpecial=True)


class FuncTask(Task):
    def __init__(self, *, priority='A', ttl=1, func, kwargs=None, format_kwargs=None, desc=None):
        super().__init__(priority=priority, ttl=ttl)

        self._func = func
        self._kwargs = {} if kwargs is None else kwargs

        self._desc = 'priority=%s, ttl=%s' % (priority, ttl)
        self._desc = '%s. %s' % (self._desc, desc or func.__name__)

        _s = (format_kwargs or format_dict)(kwargs)
        if len(_s) > 0:
            self._desc = '%s: %s' % (self._desc, _s)

    @property
    def kwargs(self):
        return self._kwargs

    def __str__(self):
        return self._desc

    def go(self):
        self._func(**self._kwargs)


class TaskScheduler:
    def __init__(self):
        self._init()

    def _init(self):
        self._q = queue.PriorityQueue()
        self._threads = []

        self._n_worker = 0
        self._n_worker_lock = threading.Lock()

        self._shutdown_event = threading.Event()

        self._failures = []
        self._failures_lock = threading.Lock()

        self._threading_local = threading.local()
        self._threading_local.task = None

    def add_task(self, task):
        self._q.put(task)

    def wait(self):
        self._q.join()

        with self._failures_lock:
            res, self._failures[:] = self._failures[:], []
        return res

    def shutdown(self):
        assert len(self._threads) == self._n_worker

        if len(self._threads) == 0:
            return

        for _ in range(len(self._threads)):
            self.add_task(_shutdown_task)

        self._shutdown_event.wait()

        for _ in self._threads:
            _.join()

        self._init()

    def _onshutdown(self):
        with self._n_worker_lock:
            self._n_worker -= 1

            if self._n_worker == 0:
                self._shutdown_event.set()

    def start(self, *, n_worker=3, WorkerFactory=threading.Thread):
        n = math.floor(math.log10(n_worker)) + 1
        for i in range(1, n_worker + 1):
            t = WorkerFactory(target=self._func_work, daemon=True, name='%%0%dd' % n % i)
            t.start()
            self._threads.append(t)

        self._n_worker = n_worker

    def register_task(self, func=None, *, FuncTaskFactory=FuncTask, **_kwargs):
        if func is None:
            return functools.partial(self.register_task, FuncTaskFactory=FuncTaskFactory, **_kwargs)

        @functools.wraps(func)
        def _func(**kwargs):
            self.add_task(FuncTaskFactory(func=func, kwargs=kwargs, **_kwargs))

        return _func

    def _add_failure(self, _failure):
        with self._failures_lock:
            self._failures.append(_failure)

    def _func_work(self):
        logging.debug('started')

        while True:
            task = self._q.get()
            self._threading_local.task = task

            if task is _shutdown_task:
                self._onshutdown()
                logging.debug('shutdowned')
                break

            try:
                logging.info('[Doing] %s' % task)
                task.run()
            except Exception:
                tb_msg = traceback.format_exc()
                if task.should_run():
                    self.add_task(task)
                    logging.warning('[Retry later] %s\n%s' % (task, tb_msg))
                else:
                    self._add_failure((task, tb_msg))
                    logging.error('[Failed] %s\n%s' % (task, tb_msg))
            else:
                logging.info('[Done] %s' % task)

            self._q.task_done()

    def current_task(self):
        return self._threading_local.task

    @property
    def d(self):
        try:
            self._threading_local.d
        except AttributeError:
            self._threading_local.d = {}

        return self._threading_local.d

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.shutdown()



import setuptools

import dl_coursera


def _read(filename):
    with open(filename, encoding='UTF-8') as ifs:
        return ifs.read()


def _readlines(filename):
    return [_ for _ in _read(filename).split('\n') if len(_) > 0]


setuptools.setup(
    name=dl_coursera.app_name,
    version=dl_coursera.app_version,
    author='Fenglei Zhang',
    author_email='fenglei4518@hotmail.com',
    description='A simple, fast, and reliable Coursera crawling & downloading tool',
    long_description=_read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/feng-lei/dl_coursera',
    keywords=['dl_coursera', 'coursera', 'coursera-dl', 'download', 'education', 'MOOC'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Education'
    ],
    python_requires='>=3',
    install_requires=_readlines('environments.txt'),
    extras_require={'dev': _readlines('environments-dev.txt')},
    py_modules=['dl_coursera_run'],
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            '%s=dl_coursera_run:main' % dl_coursera.app_name
        ]
    },
    include_package_data=True,
    package_data={
        'dl_coursera.resource': ['*.zip', 'template/*']
    },
    platforms=['any']
)
